from os.path import join, dirname, basename, normpath


class Path(object):
    def __init__(self, *paths):
        self.path = normpath(join(*paths))

    def join(self, *others):
        return Path(str(self), *others)

    def __str__(self):
        return self.path

    def rooted(self, paths):
        return map(self.join, paths)

    @property
    def parent(self):
        return self.join('..')

    @property
    def basename(self):
        return basename(str(self))

    @property
    def dirname(self):
        return dirname(str(self))



from fabric.tasks import WrappedCallableTask
from fake.api import execute, env, settings


class WrappedCallableDependenciesTask(WrappedCallableTask):
    def __init__(self, *args, **kwargs):
        super(WrappedCallableDependenciesTask, self).__init__(*args, **kwargs)
        self.prereqs = []
        self.postreqs = []

    def run(self, *args, **kwargs):
        for func in self.prereqs:
            execute(func, host=env.host)
        # set role specific variables from roledefs
        with settings(**self.role_settings()):
            result = super(WrappedCallableDependenciesTask, self).run(*args, **kwargs)
        for func in self.postreqs:
            execute(func, host=env.host)
        return result

    def role_settings(self):
        result = {'host_roles': []}
        rdefs = env.get('roledefs', {})
        for erole in env.effective_roles:
            settings = rdefs.get(erole, {})
            result.update(settings if isinstance(settings, dict) else {})
        for role in [r for r in env.roles if r in rdefs]:
            hosts = rdefs[role]
            if isinstance(hosts, dict) and 'hosts' in hosts:
                hosts = hosts['hosts']
            if callable(hosts):
                hosts = hosts()
            if isinstance(hosts, list) and env.host in hosts:
                result['host_roles'].append(role)
        for key in result.iterkeys():
            value = result[key]
            result[key] = value() if callable(value) else value
        return result


def task(*args, **kwargs):
    # if the wrapper hasn't been called
    if args and not kwargs:
        return WrappedCallableDependenciesTask(args[0])

    def wrapper(func):
        return WrappedCallableDependenciesTask(func, *args, **kwargs)

    return wrapper



class SCMStrategy(object):
    def check(self):
        """
        Check that the repository is reachable.
        """
        raise NotImplementedError

    def test(self):
        """
        Determine if the repo has been mirrored.
        """
        raise NotImplementedError

    def clone(self):
        """
        Clone the repo to the cache.
        """
        raise NotImplementedError

    def update(self):
        """
        Update the repo mirror to reflect the origin state.
        """
        raise NotImplementedError

    def release(self):
        """
        Copy repo to releases.
        """
        raise NotImplementedError

    def fetch_revision(self):
        """
        Determine the revision that will be deployed.
        """
        raise NotImplementedError



from fabric.utils import puts
from fake.scm.base import SCMStrategy
from fake.api import task, env, run, paths, cd, test


class Git(SCMStrategy):
    def git(self, cmd):
        return run("git %s" % cmd)

    def check(self):
        self.git("ls-remote --heads %(repo_url)s > /dev/null" % env)

    def test(self):
        return test("[ -f %(repo_path)s/HEAD ]" % paths)

    def clone(self):
        args = (env.repo_url, paths.repo_path)
        self.git("clone --mirror %s %s > /dev/null" % args)

    def update(self):
        self.git("remote update --prune > /dev/null")

    def release(self):
        args = (env.branch, paths.release_path)
        self.git("archive %s | tar -x -f - -C %s" % args)

    def fetch_revision(self):
        return self.git("rev-list --max-count=1 %(branch)s" % env)


# this is safe to do here, because this module will be included as necessary,
# after the env settings in the fabfile.py have been set
strategy = env.get('git_strategy', Git())


@task
def check():
    """
    Check that the repository is reachable.
    """
    strategy.check()


@task
def create_release():
    """
    Copy repo to releases.
    """
    if strategy.test():
        puts("mirror exists at %(repo_path)s" % paths)
    else:
        with cd(paths.deploy_path):
            strategy.clone()

    with cd(paths.repo_path):
        strategy.update()
        run("mkdir -p %(release_path)s" % paths)
        strategy.release()


@task
def set_current_revision():
    with cd(paths.repo_path):
        env.current_revision = strategy.fetch_revision()



# so we have parity with fabric
from fabric.api import *

from fake.state import paths
from fake.decorators import task
from fake.context_managers import cd
from fake.tasks.helpers import before, after
from fake.operations import test



from importlib import import_module
import time

from fabric.utils import puts

from fake.utils import Path
from fake.decorators import task
from fake.api import execute, env, paths, run, cd
from fake.tasks.framework import *
from fake.operations import test


# Set defaults
paths.setdefault('deploy_path', '/var/www/app')
paths.setdefault('repo_path', lambda: Path(paths.deploy_path, 'repo'))
paths.setdefault('shared_path', lambda: Path(paths.deploy_path, 'shared'))
paths.setdefault('releases_path', lambda: Path(paths.deploy_path, 'releases'))
paths.setdefault('release_path', lambda: paths.current_path)
paths.setdefault('current_path', lambda: Path(paths.deploy_path, 'current'))

env.setdefault('linked_dirs', [])
env.setdefault('linked_files', [])
env.setdefault('scm', 'git')
env.setdefault('branch', 'master')
env.setdefault('keep_releases', 5)


def _set_scm():
    """
    Import correct SCM module into environment.
    """
    g = globals()
    if 'scm' not in g:
        g['scm'] = import_module("fake.scm.%(scm)s" % env)


@task
def starting():
    """
    Start a deployment, make sure server(s) ready.
    """
    execute('check', host=env.host)
    execute('set_previous_revision', host=env.host)


@task
def check():
    """
    Check required files and directories exist
    """
    if 'deploy_path' not in env or 'repo_url' not in env:
        raise RuntimeError("You must set deploy_path and repo_url in your environment")
    _set_scm()
    execute(scm.check, host=env.host)
    run("mkdir -p %s %s" % (paths.shared_path, paths.releases_path))
    if len(env.linked_dirs) > 0:
        dirs = map(str, paths.shared_path.rooted(env.linked_dirs))
        run("mkdir -p %s" % " ".join(dirs))
    if len(env.linked_files) > 0:
        dirs = [paths.shared_path.join(f).dirname for f in env.linked_files]
        run("mkdir -p %s" % " ".join(dirs))


@task
def set_previous_revision():
    target = paths.release_path.join("REVISION")
    if test("[ -f %s ]" % target):
        env.previous_revision = run('cat %s' % target)


@task
def updating():
    """
    Update server(s) by setting up a new release.
    """
    # this may have been set on a previous run on another server.
    # it's used later, so don't update if it's already been set.
    if 'release_timestamp' not in env:
        env.release_timestamp = time.strftime('%Y%m%d%H%M%S', time.gmtime())
    paths.release_path = paths.releases_path.join(env.release_timestamp)
    _set_scm()
    execute(scm.create_release, host=env.host)
    execute(scm.set_current_revision, host=env.host)
    with cd(paths.release_path):
        run("echo \"%s\" >> REVISION" % env.current_revision)
    execute("symlink_folders", host=env.host)
    execute("symlink_files", host=env.host)


@task
def symlink_folders():
    """
    Symlink linked folders.
    """
    if len(env.linked_dirs) == 0:
        return

    # make sure parents exist in the release path
    dirs = [paths.release_path.join(d).dirname for d in env.linked_dirs]
    run("mkdir -p %s" % " ".join(dirs))

    # link some dirs
    for d in env.linked_dirs:
        target = paths.release_path.join(d)
        source = paths.shared_path.join(d)
        if test("[ -L %s ]" % target):
            continue
        if test("[ -d %s ]" % target):
            run("rm -rf %s" % target)
        run("ln -s %s %s" % (source, target))


@task
def symlink_files():
    """
    Symlink linked files.
    """
    if len(env.linked_files) == 0:
        return

    # make sure parents exist in the release path
    dirs = [paths.release_path.join(f).dirname for f in env.linked_files]
    run("mkdir -p %s" % " ".join(dirs))

    # link some files
    for f in env.linked_files:
        target = paths.release_path.join(f)
        source = paths.shared_path.join(f)
        if test("[ -L %s ]" % target):
            continue
        if test("[ -f %s ]" % target):
            run("rm -f %s" % target)
        run("ln -s %s %s" % (source, target))


@task
def publishing():
    """
    Symlink current release.
    """
    tmp_current_path = paths.release_path.parent.join(paths.current_path.basename)
    run("ln -s %s %s" % (paths.release_path, tmp_current_path))
    run("mv %s %s" % (tmp_current_path, paths.current_path.parent))


@task
def finishing():
    execute("cleanup", host=env.host)


@task
def cleanup():
    """
    Clean up old releases
    """
    puts("Keeping last %i releases" % env.keep_releases)
    releases = run("ls -xtr %(releases_path)s" % paths).split()
    if len(releases) <= env.keep_releases:
        return

    for release in releases[:(len(releases) - env.keep_releases)]:
        run("rm -rf %s" % paths.releases_path.join(release))


@task
def reverting():
    releases = run("ls -xt %s" % paths.releases_path).split()
    if len(releases) < 2:
        raise RuntimeError("Cannot rollback without a previous deploy")
    env.release_timestamp = releases[1]
    paths.release_path = paths.releases_path.join(env.release_timestamp)


@task
def finishing_rollback():
    execute("cleanup_rollback", host=env.host)


@task
def cleanup_rollback():
    """
    Remove and archive rolled-back release.
    """
    last_release = run("ls -xt %s" % paths.releases_path).split()[0]
    last_release_path = paths.releases_path.join(last_release)
    args = (paths.current_path, last_release_path)
    if test("[ `readlink %s` != %s ]" % args):
        fname = Path(paths.deploy_path, "rolled-back-release-%s.tar.gz" % last_release)
        run("tar -czf %s %s" % (fname, last_release_path))
        run("rm -rf %s" % last_release_path)
    else:
        puts("Last release is the current release, skip cleanup_rollback.")



from fabric.state import env


class CallableEnv(env.__class__):
    def __getitem__(self, key):
        # if it's set in env, then it trumps whats here
        if key in env:
            return env[key]
        value = super(CallableEnv, self).__getitem__(key)
        return value() if callable(value) else value

paths = CallableEnv()



#!/usr/bin/env python
from setuptools import setup, find_packages
from fake import __version__

setup(
    name="fake",
    version=__version__,
    description="Make Python's Fabric act like Ruby's Capistrano",
    author="Brian Muller",
    author_email="bamuller@gmail.com",
    license="MIT",
    url="http://github.com/bmuller/fake",
    packages=find_packages(),
    requires=["fabric"],
    install_requires=['fabric>=1.12']
)



"""
Based on https://github.com/capistrano/capistrano/blob/master/lib/capistrano/tasks/framework.rake
"""

from fake.decorators import task
from fake.api import execute, env


@task
def starting():
    """
    Start a deployment, make sure server(s) ready.
    """


@task
def started():
    """
    Started
    """


@task
def updating():
    """
    Update server(s) by setting up a new release.
    """


@task
def updated():
    """
    Updated
    """


@task
def reverting():
    """
    Revert server(s) to previous release.
    """


@task
def reverted():
    """
    Reverted
    """


@task
def publishing():
    """
    Publish the release.
    """


@task
def published():
    """
    Published
    """


@task
def finishing():
    """
    Finish the deployment, clean up server(s).
    """


@task
def finishing_rollback():
    """
    Finish the rollback, clean up server(s).
    """


@task
def finished():
    """
    Finished
    """


@task
def rollback():
    """
    Rollback to previous release.
    """
    tasks = (
        'starting started reverting reverted '
        'publishing published finishing_rollback finished'
    )
    for task in tasks.split():
        execute(task, host=env.host)


@task
def deploy():
    """
    Deploy a new release.
    """
    tasks = (
        'starting started updating updated '
        'publishing published finishing finished'
    )
    for task in tasks.split():
        execute(task, host=env.host)
